#######################################################################
#  Indiana University Opie Lab Software License, Version 1.2
# 
#  Copyright (C) 2002 The Trustees of Indiana University.
#  All rights reserved.
# 
#  Redistribution and use in source and binary forms, with or without
#  modification, are permitted provided that the following conditions are
#  met:
# 
#  1) All redistributions of source code must retain the above
#     copyright notice, the list of authors in the original source
#     code, this list of conditions and the disclaimer listed in this
#     license;
# 
#  2) All redistributions in binary form must reproduce the above
#     copyright notice, this list of conditions and the disclaimer
#     listed in this license in the documentation and/or other
#     materials provided with the distribution;
# 
#  3) Any documentation included with all redistributions must include
#     the following acknowledgement:
# 
#       "This product includes software developed by the Indiana
#       University Opie Lab.  For further information please visit
#       http://www.cs.indiana.edu/~dswise"
# 
#     Alternatively, this acknowledgment may appear in the software
#     itself, and wherever such third-party acknowledgments normally
#     appear.
# 
#  4) The name "Indiana University" or "Indiana University
#     Opie Lab" shall not be used to endorse or promote
#     products derived from this software without prior written
#     permission from Indiana University.  For written permission,
#     please contact http://www.cs.indiana.edu/~dswise.
# 
#  5) Products derived from this software may not use "Indiana
#     Univeristy" name nor may "Indiana University" appear in their name,
#     without prior written permission of the Indiana University.
# 
#  Indiana University provides no reassurances that the source code
#  provided does not infringe the patent or any other intellectual
#  property rights of any other entity.  Indiana University disclaims any
#  liability to any recipient for claims brought by any other entity
#  based on infringement of intellectual property rights or otherwise.
#
#  LICENSEE UNDERSTANDS THAT SOFTWARE IS PROVIDED "AS IS" FOR WHICH
#  NO WARRANTIES AS TO CAPABILITIES OR ACCURACY ARE MADE. INDIANA
#  UNIVERSITY GIVES NO WARRANTIES AND MAKES NO REPRESENTATION THAT
#  SOFTWARE IS FREE OF INFRINGEMENT OF THIRD PARTY PATENT, COPYRIGHT, OR
#  OTHER PROPRIETARY RIGHTS.  INDIANA UNIVERSITY MAKES NO WARRANTIES THAT
#  SOFTWARE IS FREE FROM "BUGS", "VIRUSES", "TROJAN HORSES", "TRAP
#  DOORS", "WORMS", OR OTHER HARMFUL CODE.  LICENSEE ASSUMES THE ENTIRE
#  RISK AS TO THE PERFORMANCE OF SOFTWARE AND/OR ASSOCIATED MATERIALS,
#  AND TO THE PERFORMANCE AND VALIDITY OF INFORMATION GENERATED USING
#  SOFTWARE.
###################################################################

####################################################################
#  cil_r.pl
#    This is a reader for cil files that have been generated with
#    the Edison Design Group software and then run through the
#    edgcpdisp utility for conversion to a human readable format.
#
#    The file will be read and then displayed to stdout in a format
#    that is Scheme-friendly (lists with parens).
#
#    If you are about to try and modify this then you should be
#    forewarned that a good knowledge of regular expressions is
#    required as they are used extensively.  For this I reccomend
#    any decent Perl book or a look through "man egrep".
#####################################################################

#####################################################################
# These variables contain the default values for the array
# representations of certain types of nodes.  A tilde ("~")
# represents an unitialized value.
#####################################################################

#####################################################################
# Things to fix/unresolved issues:
#   multi-line fields
#   multiple instances of a field name
#   pragma entities
#   pragma values
#   assoc_scope with routines
#   are bools always FALSE if undefined?
#   print (0 0 0) or NULL?
#####################################################################

#####################################################################
# Things that I think are completely done:
#   il_header
#   source-file's
#   scope's
#   type's
#   variable's
#   
#####################################################################

@il_header 		        = ("~","~","~","~","~","~","~","~","~","~","~","~","~","~","~");
$undefined_string               = "~";
$line_no = 0;
$current_mem_region		= -1;
$found = 0;
$address_mode = 2;

#######################################################################
# quote receives a string and returns it with quotes on each side.
# i.e. with the argument: a string, the return is: "a string".
#######################################################################
sub quote {
    $arg1 = shift(@_);
    return "\"$arg1\"";
}

sub pair {
    $arg1 = shift(@_);
    $arg2 = shift(@_);
    return "($arg1 $arg2)";
}

sub triple {
    $arg1 = shift(@_);
    $arg2 = shift(@_);
    $arg3 = shift(@_);
    return "($arg1 $arg2 $arg3)";
}

sub new_blank_list {
    $arg1 = shift(@_);
    $ls = [($undefined_string)];

    for($i=0; $i<$arg1; $i++) {
	$$ls[$i] = $undefined_string;
    }

    return $ls;
}

sub parse_address {
    $ref = shift(@_);
    $index = shift(@_);
    $string = shift(@_);
    $field_name = shift(@_);
    $node_type_name = shift(@_);

    if($found) {
	return;
    }

    if( !($string =~ /^$field_name:\s*./) ) {
	return;
    }

    if( $string =~ /^$field_name:\s*([\w|-]*)\s.*$node_type_name.(\d+).*$/ ) {
	if($address_mode == 0) {
	    $addr = $2;
	}
	elsif($address_mode == 1) {
	    $addr = pair($1, $2);
	}
	else {
	    $addr = triple($1, $node_type_name, $2);
	}
    }	  
    elsif ($string =~ /^$field_name:.*NULL.*$/) {
	if($address_mode == 0) {
	    $addr = "0";
	}
	elsif ($address_mode == 1) {
	    $addr = "(0 0)";
	}
	else {
	    $addr = "(0 0 0)";
	}
    }
    else {
	printf STDERR "Warning: parsing an unknown two line address, line no. %d.\n", $line_no;
	$s = $argv;
	$line_no++;

	if( $s =~ /^\s*([\w|-]*)\s.*$node_type_name.(\d+).*$/ ) {
	    if($address_mode == 0) {
		$addr = $2;
	    }
	    elsif($address_mode == 1) {
		$addr = pair($1, $2);
	    }
	    else {
		$addr = triple($1, $node_type_name, $2);
	    }
	}	  
	elsif ($s =~ /.*NULL.*$/) { 
	    if($address_mode == 0) {
		$addr = "0";
	    }
	    elsif ($address_mode == 1) {
		$addr = "(0 0)";
	    }
	    else {
		$addr = "(0 0 0)";
	    }
	}
	else {
	    printf STDERR "Irrecoverable error at line number %d:\n", $line_no;
	    printf STDERR "  invalid %s address, failure reading a second line.\n", $field_name;
	    printf STDERR "  line 1: %s", $string;
	    printf STDERR "  line 2: %s\n", $s;
	    exit;
	}
    }

    $found = 1;
    $$ref[$index] = $addr;

    return;
}

sub parse_address_two_lines {
    $ref = shift(@_);
    $index = shift(@_);
    $string = shift(@_);
    $field_name = shift(@_);
    $node_type_name = shift(@_);

    if($found) {
	return;
    }

    if( !($string =~ /^$field_name:\s*/) ) {
	return;
    }

    $s = <ARGV>;
    $line_no++;

    if( $s =~ /^\s*([\w|-]*)\s.*$node_type_name.(\d+).*$/ ) {
	if($address_mode == 0) {
	    $addr = $2;
	}
	elsif($address_mode == 1) {
	    $addr = pair($1, $2);
	}
	else {
	    $addr = triple($1, $node_type_name, $2);
	}
    }	  
    else {
	if($address_mode == 0) {
	    $addr = "0";
	}
	elsif ($address_mode == 1) {
	    $addr = "(0 0)";
	}
	else {
	    $addr = "(0 0 0)";
	}
    }

    $found = 1;
    $$ref[$index] = $addr;

    return;
}

sub parse_id {
    $ref = shift(@_);
    $index = shift(@_);
    $string = shift(@_);
    $field_name = shift(@_);

    if($found) {
	return;
    }

    if( !($string =~ /^$field_name:\s*\S*\s.*id-name.\d+: \"(.*)\"$/ ) ) {
	return;
    }	  

    $found = 1;
    $$ref[$index] = $1;

    return;

}

sub parse_other {
    $ref = shift(@_);
    $index = shift(@_);
    $string = shift(@_);
    $field_name = shift(@_);

    if($found) {
	return;
    }

    if( $string =~ /^$field_name:\s*NULL\s*$/ ) {	
	$found = 1;
	$$ref[$index] = "NULL";

	return;
    }

    if( !($string =~ /^$field_name:.*other-text.\d+: \"(.*)\"$/ ) ) {
	return;
    }	  

    $found = 1;
    $$ref[$index] = quote($1);

    return;

}

sub parse_string {
    $ref = shift(@_);
    $index = shift(@_);
    $string = shift(@_);
    $field_name = shift(@_);

    if($found) {
	return;
    }

    if( !($string =~ /^$field_name:\s*(\S.*)$/) ) {
	return;
    }

    $found = 1;
    $$ref[$index] = quote($1);

    return;
}

sub parse_string_no_quotes {
    $ref = shift(@_);
    $index = shift(@_);
    $string = shift(@_);
    $field_name = shift(@_);

    if($found) {
	return;
    }

    if( !($string =~ /^$field_name:\s*(\S.*)$/) ) {
	return;
    }

    $found = 1;
    $$ref[$index] = $1;
    return;
}

sub parse_ignore {
    $string = shift(@_);
    $field_name = shift(@_);
    
    if( $found ) {
	return;
    }

    if( $string =~ /^$field_name:/ ) {
	$found = 1;	
    }

    return;
}

sub parse_ignore_wentry {
    $string = shift(@_);
    $field_name = shift(@_);
    
    if( $found ) {
	return;
    }

    if( $string =~ /^$field_name:/ ) {
	$found = 1;	
	if( $string =~ /^$field_name:\s*$/ ) {
	    $string = <ARGV>;
	    $line_no++;
	}
    }

    return;
}

sub parse_bool {
    $ref = shift(@_);    
    $index = shift(@_);
    $string = shift(@_);
    $field_name = shift(@_);

    if( $found ) {
	return;
    }

    if( !($string =~ /^$field_name:/ )) {
	return;
    }

    $found = 1;

    if( ($string =~ /^$field_name:\s*(\S.*)$/) ) {	
	$s = $1;
    }
    else {
	printf STDERR "Warning: unknown double bool line, line no. %d\n", $line_no;
	$s = <ARGV>;
	$line_no++;
    }

    if(!($s =~ /TRUE|FALSE/)) {
	printf STDERR "Found a malformed bool, line number %d\n", $line_no;
    }
    
    $$ref[$index] = $s;
    return;
}    

sub parse_bool_two_lines {
    $ref = shift(@_);    
    $index = shift(@_);
    $string = shift(@_);
    $field_name = shift(@_);

    if( $found ) {
	return;
    }

    if( !($string =~ /^$field_name:\s*$/) ) {
	return;
    }
 
    $found = 1;
    $s = <ARGV>;
    $line_no++;
    
    if(!($s =~ /\s*(TRUE|FALSE)/)) {
	printf STDERR "Found a malformed bool (two line boolean), line number %d\n", $line_no;
    }

    $$ref[$index] = $1;
    return;
}

sub print_region {

   

    if($file_scope) {
	printf("  ( ");
	for($i = 0; $i<scalar(@il_header); $i++ ) {
	    printf("%s ", $il_header[$i]);
	}
	printf(")\n");
    }

    printf("  ( %d %s\n", $current_mem_region_no, $current_mem_region_name);

    print("    (\n");
    for($i=0; $i < $current_file; $i++) {
	$p = $files[$i];
	print("      ( ");
	for($j = 0; $j<scalar(@$p); $j++) {
	    print($$p[$j]);
	    print(" ");
	}    
	print(")\n");
    }
    print("    )\n");

    print("    (\n");
    for($i=0; $i < $current_scope; $i++) {
	$p = $scopes[$i];
	print("      ( ");
	for($j = 0; $j<scalar(@$p); $j++) {
	    print($$p[$j]);
	    print(" ");
	}    
	print(")\n");
    }
    print("    )\n");

    print("    (\n");
    for($i=0; $i < $current_pragma; $i++) {
	$p = $pragmas[$i];
	print("      ( ");
	for($j = 0; $j<scalar(@$p); $j++) {
	    print($$p[$j]);
	    print(" ");
	}    
	print(")\n");
    }
    print("    )\n");

    print("    (\n");
    for($i=0; $i < $current_type; $i++) {
	$p = $types[$i];
	print("      ( ");
	for($j = 0; $j<scalar(@$p); $j++) {
	    print($$p[$j]);
	    print(" ");
	}    
	print(")\n");
    }
    print("    )\n");

    print("    (\n");
    for($i=0; $i < $current_var; $i++) {
	$p = $vars[$i];
	print("      ( ");
	for($j = 0; $j<scalar(@$p); $j++) {
	    print($$p[$j]);
	    print(" ");
	}    
	print(")\n");
    }
    print("    )\n");    

    print("    (\n");
    for($i=0; $i < $current_routine; $i++) {
	$p = $routines[$i];
	print("      ( ");
	for($j = 0; $j<scalar(@$p); $j++) {
	    print($$p[$j]);

	    print(" ");
	}    
	print(")\n");
    }
    print("    )\n");    

    print("    (\n");
    for($i=0; $i < $current_field; $i++) {
	$p = $fields[$i];
	print("      ( ");
	for($j = 0; $j<scalar(@$p); $j++) {
	    print($$p[$j]);
	    print(" ");
	}    
	print(")\n");
    }
    print("    )\n");

    print("    (\n");
    for($i=0; $i < $current_dynamic; $i++) {
	$p = $dynamics[$i];
	print("      ( ");
	for($j = 0; $j<scalar(@$p); $j++) {
	    print($$p[$j]);
	    print(" ");
	}    
	print(")\n");
    }
    print("    )\n");

    print("    (\n");
    for($i=0; $i < $current_param; $i++) {
	$p = $params[$i];
	print("      ( ");
	for($j = 0; $j<scalar(@$p); $j++) {
	    print($$p[$j]);
	    print(" ");
	}    
	print(")\n");
    }
    print("    )\n");    

    print("    (\n");
    for($i=0; $i < $current_constant; $i++) {
	$p = $constants[$i];
	print("      ( ");
	for($j = 0; $j<scalar(@$p); $j++) {
	    print($$p[$j]);
	    print(" ");
	}    
	print(")\n");
    }
    print("    )\n");

    print("    (\n");
    for($i=0; $i < $current_statement; $i++) {
	$p = $statements[$i];
	print("      ( ");
	for($j = 0; $j<scalar(@$p); $j++) {
	    print($$p[$j]);
	    print(" ");
	}    
	print(")\n");
    }
    print("    )\n");    
 
    print("    (\n");
    for($i=0; $i < $current_expr; $i++) {
	$p = $exprs[$i];
	print("      ( ");
	for($j = 0; $j<scalar(@$p); $j++) {
	    print($$p[$j]);
	    print(" ");
	}    
	print(")\n");
    }
    print("    )\n"); 

    print("    (\n");
    for($i=0; $i < $current_label; $i++) {
	$p = $labels[$i];
	print("      ( ");
	for($j = 0; $j<scalar(@$p); $j++) {
	    print($$p[$j]);
	    print(" ");
	}    
	print(")\n");
    }
    print("    )\n");
    print("  )\n");
}    


print("(\n");

while(<ARGV>) {
    $line_no++;

    if($_ =~ /^Intermediate language for memory region /) {
	$temp = $_;

	if(defined($current_mem_region_no)) {
	    print_region();
	}
	if($temp =~ /^Intermediate language for memory region (\d*) \(function "(.*)"\):$/) {
	    $file_scope = 0;
	    $current_mem_region_no = int($1);
	    $current_mem_region_name = $2;
	}
	elsif ($temp =~ /^Intermediate language for memory region (\d*) \((.*)\):$/) {
	    $file_scope = 1;
	    $current_mem_region_no = int($1);
	    $current_mem_region_name = "file-scope";
	}
	$current_var = 0;
	$current_type = 0;
	$current_scope = 0;
	$current_expr = 0;
	$current_statement = 0;
	$current_scope = 0;
	$current_routine = 0;
	$current_param = 0;
	$current_file = 0;
	$current_pragma = 0;
	$current_dynamic = 0;
	$current_label = 0;
	$current_field = 0;
        $current_constant = 0;
    }
    elsif( $_ =~ /^il_header:$/) {
	while(<ARGV>) {
	    $line_no++;
	    $found = 0;
	    $current_entry = 0;

	    if( $_ =~ /^$/ ) {
		last;
	    }
	    $bah = 0;
	    parse_address          (\@il_header, 0, $_, "primary_source_file", "source-file");
	    parse_address          (\@il_header, 1, $_, "primary_scope", "scope");
	    parse_address          (\@il_header, 2, $_, "main_routine", "routine");
	    parse_other            (\@il_header, 3, $_, "compiler_version" );
	    parse_other            (\@il_header, 4, $_, "time_of_compilation" ); 
	    parse_bool             (\@il_header, 5, $_, "plain_chars_are_signed" );
	    parse_string           (\@il_header, 6, $_, "source_language" );
	    parse_bool             (\@il_header, 7, $_, "pcc_compatibility_mode" );
	    parse_bool             (\@il_header, 8, $_, "enum_type_is_integral" );
	    parse_string_no_quotes (\@il_header, 9, $_, "default_max_member_alignment");
	    parse_bool             (\@il_header,10, $_, "near_and_far_are_enabled");
	    parse_bool             (\@il_header,11, $_, "far_data_pointers");
	    parse_bool             (\@il_header,12, $_, "far_code_pointers");
	    parse_bool             (\@il_header,13, $_, "UCN_identifiers_used" );
	    parse_address_two_lines(\@il_header,14, $_, "nontag_types_used_in_exception_or_rtti","unknown" );

	    if(!$found) {
                printf STDERR "Unrecongized line, reading an il_entry, line %d:\n", $line_no;
		printf STDERR "  %s\n", $_;
	    }
	}
    }
    elsif( $_ =~ /^([\w|-]+)[ ]+variable.(\d+)$/) {	
	$new_var = new_blank_list(30);
	$vars[$current_var] = $new_var;
	$$new_var[0] = $2;

	while(<ARGV>) {
            $line_no++;
	    $found = 0;

	    if( $_ =~ /^$/ ) {
		last;
	    }

	    parse_ignore(           $_,              "source_corresp");
	    
	    parse_address(          $new_var,  1, $_,"next", "variable");

	    # source correspondence stuff
	    parse_id(               $new_var,  2, $_,"  name");
	    parse_id(               $new_var,  3, $_,"  unmangled_name");
	    parse_string_no_quotes( $new_var,  4, $_,"  decl_position.seq");
	    parse_string_no_quotes( $new_var,  5, $_,"  decl_position.column");
	    parse_bool(             $new_var,  6, $_,"  referenced");
	    parse_bool(             $new_var,  7, $_,"  needed");
	    parse_bool(             $new_var,  8, $_,"  is_local_to_function");
	    parse_string          ( $new_var,  9, $_,"  name_linkage"); 
	    parse_bool(             $new_var, 10, $_,"  has_associated_pragma");
	    parse_bool(             $new_var, 11, $_,"  static_used_by_instantiation");

	    parse_address(          $new_var, 12, $_,"type", "type");
	    parse_address(          $new_var, 13, $_,"assoc_param_type", "param-type");
	    parse_string_no_quotes( $new_var, 14, $_,"storage_class");
	    parse_string_no_quotes( $new_var, 15, $_,"declared_storage_class");
	    parse_string(           $new_var, 16, $_,"init_kind");
	    parse_bool(             $new_var, 17, $_,"address_taken");
	    parse_address(          $new_var, 18, $_,"dynamic", "dynamic-init"); 
	    parse_bool_two_lines(   $new_var, 19, $_,"param_value_has_been_changed");
	    parse_bool_two_lines(   $new_var, 20, $_,"param_used_more_than_once"); 
	    parse_bool(             $new_var, 21, $_,"is_parameter");	
	    parse_bool(             $new_var, 22, $_,"is_handler_param");	
	    parse_bool(             $new_var, 23, $_,"referenced_non_locally");	 
	    parse_bool(             $new_var, 24, $_,"is_partially_initialized");	
	    parse_bool(             $new_var, 25, $_,"superseded_external");	
	    parse_bool(             $new_var, 26, $_,"is_compound_literal");	
	    parse_bool(             $new_var, 27, $_,"has_parenthesized_initializer");
	    parse_address(          $new_var, 28, $_,"declared_type", "type");
            parse_address(          $new_var, 29, $_,"constant", "constant");
	    
	    if(!$found) {
		printf STDERR "Unrecognized line, reading a variable, line number %d\n", $line_no, $1;
                printf STDERR "  %s\n", $_;
            }	
	}
	$current_var += 1;
    }
    elsif( $_ =~ /^([\w|-]+)[ ]+routine.(\d+)$/) {
	$new_routine = new_blank_list(18);
	$routines[$current_routine] = $new_routine;
	$$new_routine[0] = $2;

	while(<ARGV>) {
            $line_no++;
	    $found = 0;

	    if( $_ =~ /^$/ ) {
		last;
	    }

	    parse_ignore(           $_,                  "source_corresp");	    
	    parse_address(          $new_routine, 1, $_, "next", "routine");
	    parse_id(               $new_routine, 2, $_, "  name");
	    parse_id(               $new_routine, 3, $_, "  unmangled_name");
	    parse_string_no_quotes( $new_routine, 4, $_, "  decl_position.seq");
	    parse_string_no_quotes( $new_routine, 5, $_, "  decl_position.column");
	    parse_bool(             $new_routine, 6, $_, "  referenced");
	    parse_bool(             $new_routine, 7, $_, "  needed");
	    parse_bool(             $new_routine, 8, $_, "  is_local_to_function");
	    parse_string(           $new_routine, 9, $_, "  name_linkage");
	    parse_address(          $new_routine, 10, $_, "type", "type");
	    parse_string_no_quotes( $new_routine, 11, $_, "storage_class");
	    parse_bool(             $new_routine, 12, $_, "defined");
	    parse_bool(             $new_routine, 13, $_, "is_called");			
	    parse_bool(             $new_routine, 14, $_, "definition_needed");
	    parse_string_no_quotes( $new_routine, 15, $_, "assoc_scope");
	    parse_bool(             $new_routine, 16, $_, "compiler_generated");
	    parse_bool(             $new_routine, 17, $_, "called");

	    if(!$found) {
		printf STDERR "Unrecognized line, reading a routine, line number %d\n", $line_no, $1;
                printf STDERR "  %s\n", $_;
            }	
	}
	$current_routine += 1;
    }
    elsif( $_ =~ /^([\w|-]+)[ ]+param-type.(\d+)$/) {
	$new_param = new_blank_list(3);
	$params[$current_param] = $new_param;
	$$new_param[0] = $2;

	while(<ARGV>) {
            $line_no++;
	    $found = 0;

	    if( $_ =~ /^$/ ) {
		last;
	    }

	    parse_address($new_param, 1, $_, "next", "param-type");
	    parse_address($new_param, 2, $_, "type", "type");

	    if(!$found) {
		printf STDERR "Unrecognized line, reading a param-type, line number %d\n", $line_no, $1;
                printf STDERR "  %s\n", $_;
            }	
	}
	$current_param += 1;
    }
    elsif( $_ =~ /^([\w|-]+)[ ]+type.(\d+)$/) {
	$new_type = new_blank_list(37);
	$types[$current_type] = $new_type;
	$$new_type[0] = $2;

	while(<ARGV>) {
            $line_no++;
	    $found = 0;

	    if( $_ =~ /^$/ ) {
		last;
	    }
	    parse_ignore(           $_,               "source_corresp");
	    parse_ignore(           $_,               "based_types" );
	    parse_ignore(           $_,               "is_class_member" );
	    parse_ignore(           $_,               "scope_depth" );
	    parse_ignore(           $_,               "source_sequence_entry" );
	    parse_ignore(           $_,               "exception_specification" );
	    parse_ignore(           $_,               "surrounding_name_linkage" );
	    parse_address(          $new_type,  1, $_,"next","type");
	    parse_string_no_quotes( $new_type,  2, $_,"kind");
	    parse_string_no_quotes( $new_type,  3, $_,"size");
	    parse_string_no_quotes( $new_type,  4, $_,"alignment");
	    parse_bool(             $new_type,  5, $_,"used_in_exception_or_rtti");
	    parse_bool(             $new_type,  6, $_,"declared_in_function_prototype");
	    parse_bool(             $new_type,  7, $_,"is_builtin_va_list");
	    parse_address(          $new_type,  8, $_,"typeinfo_var","variable");
	    # source correspondence stuff
	    parse_id(               $new_type,  9, $_,"  name");
	    parse_id(               $new_type, 10, $_,"  unmangled_name");
	    parse_string_no_quotes( $new_type, 11, $_,"  decl_position.seq");
	    parse_string_no_quotes( $new_type, 12, $_,"  decl_position.column");
	    parse_bool(             $new_type, 13, $_,"  referenced");
	    parse_bool(             $new_type, 14, $_,"  needed");
	    parse_bool(             $new_type, 15, $_,"  is_local_to_function");
	    parse_string          ( $new_type, 16, $_,"  name_linkage"); 
	    parse_bool(             $new_type, 17, $_,"  has_associated_pragma");
	    parse_bool(             $new_type, 18, $_,"  static_used_by_instantiation");

	    # based types stuff
	    parse_address(          $new_type, 19, $_,"  reference","type");
	    parse_address(          $new_type, 20, $_,"  pointer", "type");
	    parse_address(          $new_type, 21, $_,"  qualified", "type");

	    if($found) {
		# don't try to do type specific stuff
	    }

	    # int only
	    elsif($$new_type[2] =~ "tk_integer") {
		parse_string(           $new_type, 22, $_,"int_kind");
		parse_bool(             $new_type, 23, $_,"explicitly_signed");
		parse_address(          $new_type, 24, $_,"enum_info.constant_list","constant");
		parse_address(          $new_type, 25, $_,"enum_info.affiliated_type","type");
		parse_bool(             $new_type, 26, $_,"bool_type");
	    }

	    # float only
	    elsif($$new_type[2] =~ "tk_float") {
		parse_string(           $new_type, 22, $_,"float_kind");
	    }

	    # pointer only
	    elsif($$new_type[2] =~ "tk_pointer") {
		parse_address(          $new_type, 22, $_,"type_pointed_to", "type");
		parse_bool(             $new_type, 23, $_,"is_reference");
	    }

	    # routine only
	    elsif($$new_type[2] =~ "tk_routine") {
		parse_address(          $new_type, 22, $_,"return_type", "type");
		parse_address(          $new_type, 23, $_,"param_type_list","param-type");
		parse_address(          $new_type, 24, $_,"assoc_routine","routine");
		parse_bool(             $new_type, 25, $_,"has_ellipsis");
		parse_bool(             $new_type, 26, $_,"prototyped");
		parse_string(           $new_type, 27, $_,"routine_name_linkage");
		parse_bool_two_lines(   $new_type, 28, $_,"routine_name_linkage_is_explicit");
		parse_string_no_quotes( $new_type, 29, $_,"lint_argsused_flag");
		parse_bool(             $new_type, 30, $_,"value_returned_by_cctor");
		parse_bool(             $new_type, 31, $_,"assoc_routine_is_ctor");
		parse_bool(             $new_type, 32, $_,"assoc_routine_is_dtor");
		parse_string_no_quotes( $new_type, 33, $_,"lint_varargs_count");
		parse_string(           $new_type, 34, $_,"arg_pragma");
		parse_string(           $new_type, 35, $_,"qualifiers");
		parse_address(          $new_type, 36, $_,"prototype_scope","scope");
	    }

	    # array only
	    elsif ($$new_type[2] =~ "tk_array") {
		parse_address(          $new_type, 22, $_,"element_type","type");
		parse_bool(             $new_type, 23, $_,"is_static"); 
		parse_bool(             $new_type, 24, $_,"is_variable_size_array");
		parse_bool(             $new_type, 25, $_,"is_vla");
		parse_bool(             $new_type, 26, $_,"has_assoc_vla_dimension");
		parse_address(          $new_type, 27, $_,"element_count_expr","expr-node");
		parse_string_no_quotes( $new_type, 28, $_,"number_of_elements");
		parse_address(          $new_type, 29, $_,"bound_constant","constant");
	    }

	    # struct, union, class only
	    elsif ($$new_type[2] =~ "tk_struct" || $$new_type[2] =~ "tk_union") {
		parse_address(          $new_type, 22, $_,"field_list","field");
		parse_bool(             $new_type, 23, $_,"any_const_member");
		parse_bool(             $new_type, 24, $_,"any_mutable_member");
		parse_bool(             $new_type, 25, $_,"abstract");
		parse_bool(             $new_type, 26, $_,"definition_needed");
		parse_string_no_quotes( $new_type, 27, $_,"max_member_alignment");
	    }

	    # typeref only
	    elsif ($$new_type[2] =~ "tk_typeref") {
		parse_address(          $new_type, 22, $_,"typeref_type","type");
		parse_string(           $new_type, 23, $_,"qualifiers");
		parse_bool(             $new_type, 24, $_,"explicit_memory_attribute_nade_explicit");
	    }
	    
	    # ptr_to_member stuff, probably not needed
	    elsif ($$new_type[2] =~ "tk_ptr_to_member") {
		parse_address(          $new_type, 22, $_,"class_of_which_a_member","type");
	    }
	    else {	
                printf STDERR "Unrecognized type kind %s, line number %d\n", $$new_type[2], $line_no;
                printf STDERR "  %s\n", $_;
	    }
	    parse_address(          $new_type, 31, $_,"extra_info","class-type-supplement");    
	    parse_bool(             $new_type, 32, $_,"wchar_t_type");

	    if(!$found) {
                printf STDERR "Unrecognized line, reading a type, line number %d\n", $line_no, $1;
                printf STDERR "  %s\n", $_;
            }
	    
	}	      
	$current_type += 1;	      
    }
    elsif( $_ =~ /^([\w|-]+)[ ]+expr-node.(\d+)$/) {
	$new_expr = new_blank_list(12);
	$exprs[$current_expr] = $new_expr;
	$$new_expr[0] = $2;

	while(<ARGV>) {
            $line_no++;
	    $found = 0;

	    if( $_ =~ /^$/ ) {
		last;
	    }
	    parse_ignore(           $_, "source_corresp");
	    parse_address(          $new_expr,  1, $_, "next", "expr-node"); 
	    parse_string_no_quotes( $new_expr,  2, $_, "kind");
	    parse_address(          $new_expr,  5, $_,"type","type");
	    parse_string_no_quotes( $new_expr,  3, $_, "operation.kind");
	    parse_address(          $new_expr,  4, $_, "routine", "routine");
	    parse_address(          $new_expr,  6, $_, "operands", "expr-node");
	    parse_address(          $new_expr,  7, $_, "variable", "variable");
	    parse_address(          $new_expr,  8, $_, "constant", "constant"); 
	    parse_bool(             $new_expr,  9, $_, "result_is_not_used");
	    parse_bool(             $new_expr, 10, $_, "compiler_generated");
	    parse_address(          $new_expr, 11, $_, "field", "field");

       	    if(!$found) {
                printf STDERR "Unrecognized line, reading an expr-node, line number %d\n", $line_no, $1;
                printf STDERR "  %s\n", $_;
            }
                           
	}	      
	$current_expr += 1;	      
    }
    elsif( $_ =~ /^([\w|-]+)[ ]+statement.(\d+)$/) {
	$new_statement = new_blank_list(18);
	$statements[$current_statement] = $new_statement;
	$$new_statement[0] = $2;

	while(<ARGV>) {
            $line_no++;
	    $found = 0;

	    if( $_ =~ /^$/ ) {
		last;
	    }
	    parse_ignore(           $_, "source_corresp");
	    parse_address(          $new_statement,  1, $_,"next","statement");
	    parse_string_no_quotes( $new_statement,  2, $_,"kind");
	    parse_string_no_quotes( $new_statement,  3, $_,"position");    
	    parse_string_no_quotes( $new_statement,  4, $_,"final_position");
	    parse_address(          $new_statement,  5, $_,"expr","expr-node"); 
	    parse_address(          $new_statement,  6, $_,"statements","statement");
	    parse_address(          $new_statement,  7, $_,"assoc_scope","scope");
	    parse_address(           $new_statement, 8, $_,"lifetime","lifetime");
	    parse_bool(             $new_statement,  9, $_,"end_of_block_reachable");	  
	    parse_address(          $new_statement, 10, $_,"then_statement","statement");
	    parse_address(          $new_statement, 11, $_,"else_statement","statement");
	    parse_address(          $new_statement, 12, $_,"initialization", "statement");
	    parse_address(          $new_statement, 13, $_,"statement", "statement");	    
	    parse_address(          $new_statement, 14, $_,"increment", "expr-node"); 
	    parse_bool(             $new_statement, 15, $_,"has_empty_else_clause");
	    parse_address(          $new_statement, 16, $_,"dynamic_init", "dynamic-init"); 
	    parse_address(          $new_statement, 17, $_,"label", "label");
	   
       	    if(!$found) {
                printf STDERR "Unrecognized line, reading a statement, line number %d\n", $line_no, $1;
                printf STDERR "  %s\n", $_;
            }
                           
	}	      
	$current_statement += 1;	      
    }
    elsif( $_ =~ /^([\w|-]+)[ ]+constant.(\d+)$/) {
	$new_constant = new_blank_list(24);
	$constants[$current_constant] = $new_constant;
	$$new_constant[0] = $2;

	while(<ARGV>) {
            $line_no++;
	    $found = 0;

	    if( $_ =~ /^$/ ) {
		last;
	    }
	    parse_ignore(           $_,                   "source_corresp");
	    parse_address(          $new_constant,  1, $_,"next","constant");
	    
            # source correspondence stuff
	    parse_id(               $new_constant,  2, $_,"  name");
	    parse_id(               $new_constant,  3, $_,"  unmangled_name");
	    parse_string_no_quotes( $new_constant,  4, $_,"  decl_position.seq");
	    parse_string_no_quotes( $new_constant,  5, $_,"  decl_position.column");
	    parse_bool(             $new_constant,  6, $_,"  referenced");
	    parse_bool(             $new_constant,  7, $_,"  needed");
	    parse_bool(             $new_constant,  8, $_,"  is_local_to_function");
	    parse_string          ( $new_constant,  9, $_,"  name_linkage"); 
	    parse_bool(             $new_constant, 10, $_,"  has_associated_pragma");
	    parse_bool(             $new_constant, 11, $_,"  static_used_by_instantiation");

	    parse_address(          $new_constant, 12, $_,"type","type");
	    parse_address(          $new_constant, 13, $_,"expr","expr-node");
	    parse_bool(             $new_constant, 14, $_,"implicit_cast");
	    parse_bool(             $new_constant, 15, $_,"explicit_cast_applied");
	    parse_bool(             $new_constant, 16, $_,"is_reinterpret_cast");
	    parse_bool(             $new_constant, 17, $_,"non_arithmetic");
 	    parse_bool(             $new_constant, 18, $_,"is_simple_zero"); 
	    parse_bool_two_lines  ( $new_constant, 19, $_,"null_pointer_constant_ruled_out");  
	    parse_string_no_quotes( $new_constant, 20, $_,"kind"); 

	    # integer stuff
	      parse_string_no_quotes( $new_constant,  21, $_,"integer_value");
	    
	    # string stuff
	    parse_string_no_quotes( $new_constant,  21, $_,"length");
	    parse_string_no_quotes( $new_constant,  22, $_,"value");
	    
	    # float stuff	      
	    parse_string_no_quotes( $new_constant, 21, $_,"float_value");

	    # address stuff
	    parse_string_no_quotes( $new_constant, 21, $_,"address.kind");
	    parse_string_no_quotes( $new_constant, 22, $_,"address.offset");
	    
            # address.kind == abk_routine
	    parse_address(          $new_constant, 23, $_,"routine","routine");

	    # address.kind == abk_variable   
	    parse_address(          $new_constant, 23, $_,"variable","variable");
	    
	    # address.kind == abk_constant
	    parse_address(          $new_constant, 23, $_,"constant","constant");  

	    # address.kind == abk_uuidof	    
	    parse_address(          $new_constant, 23, $_,"type","type"); 

	    # dynamic init, ptr_to_member, stack_offset are ignored for now
	    
	    # aggregate
	    parse_address(          $new_constant, 21, $_,"first_constant","constant");
	    parse_address(          $new_constant, 22, $_,"last_constant","constant");    
	  
	    # init_repeat, designator and template_param are ignored for now
	    
       	    if(!$found) {
                printf STDERR "Unrecognized line, reading a constant, line number %d\n", $line_no, $1;
                printf STDERR "  %s\n", $_;
            }
                           
	}	      
	$current_constant += 1;	      
    }
    elsif( $_ =~ /^([\w|-]+)[ ]+scope.(\d+)$/) {
	$new_scope = new_blank_list(18);
	$scopes[$current_scope] = $new_scope;
	$$new_scope[0] = $2;
	
	while(<ARGV>) {
            $line_no++;
	    $found = 0;

	    if( $_ =~ /^$/ ) {
		last;
	    }

	    # unsupported stuff and header stuff
	    parse_ignore_wentry(    $_,               "namespaces");
	    parse_ignore_wentry(    $_,               "assoc_namespace");
	    parse_ignore_wentry(    $_,               "hidden names");
	    parse_ignore_wentry(    $_,               "templates");
	    parse_ignore_wentry(    $_,               "assoc_handler");
	    parse_ignore_wentry(    $_,               "local_static_variable_inits");
	    parse_ignore_wentry(    $_,               "constructor_inits");
	    parse_ignore_wentry(    $_,               "lifetime_of_local_static_vars");
	    parse_ignore_wentry(    $_,               "this_param_variable");
	    parse_ignore_wentry(    $_,               "local_static_variable_inits");

	    parse_address(          $new_scope, 1, $_,"next","scope");
	    parse_string_no_quotes( $new_scope, 2, $_,"kind"); 
	    parse_address(          $new_scope, 3, $_,"assoc_block","statement");
	    parse_address(          $new_scope, 4, $_,"lifetime","lifetime");
	    parse_address(          $new_scope, 5, $_,"constants","constant");
	    parse_address(          $new_scope, 6, $_,"types","type");
	    parse_address(          $new_scope, 7, $_,"variables","variable");
	    parse_address(          $new_scope, 8, $_,"nonstatic_variables","variable");
	    parse_address(          $new_scope, 9, $_,"labels","label");
	    parse_address(          $new_scope,10, $_,"routines","routine");
	    parse_address(          $new_scope,11, $_,"asm_entries","unknown");
	    parse_address(          $new_scope,12, $_,"scopes","scope");
	    parse_address(          $new_scope,13, $_,"dynamic_inits","dynamic-init");
	    parse_address(          $new_scope,14, $_,"pragmas","pragma");

	    # function stuff
	    parse_address(          $new_scope,15, $_,"routine.ptr","routine");
	    parse_address(          $new_scope,16, $_,"parameters","variable");
	    parse_address(          $new_scope,17, $_,"return_value_variable","variable");

	    # class_struct_union stuff
	    parse_address(          $new_scope,15, $_,"using_decls","using-decl");

	    # func_prototype or condition stuff
	    parse_address(          $new_scope,15, $_,"assoc_type","type");

       	    if(!$found) {
                printf STDERR "Unrecognized line, reading a scope, line number %d\n", $line_no, $1;
                printf STDERR "  %s\n", $_;
            }
	}	      
	$current_scope += 1;	      
    }
    elsif( $_ =~ /^([\w|-]+)[ ]+source-file.(\d+)$/) {
	$new_file = new_blank_list(15);
	$files[$current_file] = $new_file;
	$$new_file[0] = $2;
	
	while(<ARGV>) {
            $line_no++;
	    $found = 0;

	    if( $_ =~ /^$/ ) {
		last;
	    }

	    parse_address(          $new_file, 1, $_,"next","source-file");
	    parse_other(            $new_file, 2, $_,"file_name"); 
	    parse_other(            $new_file, 3, $_,"full_name"); 
	    parse_other(            $new_file, 4, $_,"name_as_written");
	    parse_string_no_quotes( $new_file, 5, $_,"first_seq_number");
	    parse_string_no_quotes( $new_file, 6, $_,"last_seq_number");
	    parse_string_no_quotes( $new_file, 7, $_,"first_line_number");
	    parse_address(          $new_file, 8, $_,"first_child_file","source-file");
	    parse_address(          $new_file, 9, $_,"last_child_file","source-file");
	    parse_bool(             $new_file,10, $_,"related_file_implicit_include_done");
	    parse_bool(             $new_file,11, $_,"is_include_file");
	    parse_bool_two_lines(   $new_file,12, $_,"included_by_system_include");
	    parse_bool(             $new_file,13, $_,"included_by_preinclude");
	    parse_bool(             $new_file,14, $_,"from_system_include_dir");

       	    if(!$found) {
                printf STDERR "Unrecognized line, reading a source-file, line number %d\n", $line_no, $1;
                printf STDERR "  %s\n", $_;
            }
	}	      
	$current_file += 1;	      
    }
    elsif( $_ =~ /^([\w|-]+)[ ]+pragma.(\d+)$/) {
	$new_pragma = new_blank_list(8);
	$pragmas[$current_pragma] = $new_pragma;
	$$new_pragma[0] = $2;
	
	while(<ARGV>) {
            $line_no++;
	    $found = 0;

	    if( $_ =~ /^$/ ) {
		last;
	    }


	    parse_ignore(           $_,               "source_sequence_entry");     
	    parse_ignore(           $_,               "source_corresp");
	    parse_address(          $new_pragma, 1, $_,"next","pragma");
	    parse_string_no_quotes( $new_pragma, 6, $_,"kind");
	    parse_address(          $new_pragma, 3, $_,"entity","unknown");
	    parse_string_no_quotes( $new_pragma, 4, $_,"position.seq");
	    parse_string_no_quotes( $new_pragma, 5, $_,"position.column");
	    parse_other(            $new_pragma, 6, $_,"pragma_text");
	    parse_bool(             $new_pragma, 7, $_,"  referenced");
	    parse_bool(             $new_pragma, 8, $_,"  needed");
	    parse_address(          $new_pragma, 3, $_,"type","type");
	    parse_string_no_quotes( $new_pragma, 9, $_,"length");
	    parse_string_no_quotes( $new_pragma,10, $_,"value");           
	    

       	    if(!$found) {
                printf STDERR "Unrecognized line, reading a pragma, line number %d\n", $line_no, $1;
                printf STDERR "  %s\n", $_;
            }
	}	      
	$current_pragma += 1;	      
    }
    elsif( $_ =~ /^([\w|-]+)[ ]+dynamic-init.(\d+)$/) {
	$new_dynamic = new_blank_list(5);
	$dynamics[$current_dynamic] = $new_dynamic;
	$$new_dynamic[0] = $2;
	
	while(<ARGV>) {
            $line_no++;
	    $found = 0;

	    if( $_ =~ /^$/ ) {
		last;
	    }

	    parse_ignore(           $_,                 "source_corresp");
	    parse_address(          $new_dynamic, 1, $_,"next","dynamic-init");
	    parse_string_no_quotes( $new_dynamic, 2, $_,"kind"); 
	    parse_address(          $new_dynamic, 3, $_,"variable","variable");
	    parse_address(          $new_dynamic, 4, $_,"constant","constant");

       	    if(!$found) {
                printf STDERR "Unrecognized line, reading a dynamic-init, line number %d\n", $line_no, $1;
                printf STDERR "  %s\n", $_;
            }
	}	      
	$current_dynamic += 1;	      
    }
    elsif( $_ =~ /^([\w|-]+)[ ]+field.(\d+)$/) {
	$new_field = new_blank_list(13);
	$fields[$current_field] = $new_field;
	$$new_field[0] = $2;
	
	while(<ARGV>) {
            $line_no++;
	    $found = 0;

	    if( $_ =~ /^$/ ) {
		last;
	    }
	    parse_ignore(           $_,               "source_corresp");
	    parse_ignore(           $_,               "  access");
	    parse_address(          $new_field, 1, $_,"next","field");
	    parse_id(               $new_field, 2, $_,"  name");
	    parse_string_no_quotes( $new_field, 3, $_,"  decl_position.seq");
	    parse_string_no_quotes( $new_field, 4, $_,"  decl_position.column");	    
	    parse_bool            ( $new_field, 5, $_,"  is_class_member");
	    parse_address(          $new_field, 6, $_,"  parent.class_type","type");
	    parse_bool            ( $new_field, 7, $_,"  referenced");
	    parse_bool            ( $new_field, 8, $_,"  needed");
	    parse_string_no_quotes( $new_field, 9, $_,"  name_linkage");
	    parse_address(          $new_field,10, $_,"type","type");
	    parse_string_no_quotes( $new_field,11, $_,"offset");
	    parse_address(          $new_field,12, $_,"assoc_type","type");

       	    if(!$found) {
                printf STDERR "Unrecognized line, reading a dynamic-init, line number %d\n", $line_no, $1;
                printf STDERR "  %s\n", $_;
            }
	}	      
	$current_field += 1;	      
    }
    elsif( $_ =~ /^([\w|-]+)[ ]+label.(\d+)$/) {
	$new_label = new_blank_list(12);
	$labels[$current_label] = $new_label;
	$$new_label[0] = $2;
	
	while(<ARGV>) {
            $line_no++;
	    $found = 0;

	    if( $_ =~ /^$/ ) {
		last;
	    }
	    parse_ignore(           $_,               "source_corresp");
	    parse_address(          $new_label, 1, $_,"next","label");
            parse_id(               $new_label, 2, $_,"  name");
	    parse_string_no_quotes( $new_label, 3, $_,"  decl_position.seq");
	    parse_string_no_quotes( $new_label, 4, $_,"  decl_position.column");
	    parse_bool            ( $new_label, 5, $_,"  referenced");	    
	    parse_bool            ( $new_label, 6, $_,"  needed");
	    parse_bool(             $new_label, 7, $_,"  is_local_to_function");
	    parse_string_no_quotes( $new_label, 8, $_,"  name_linkage");
	    parse_bool_two_lines(   $new_label, 9, $_,"reachable_by_fall_through");
	    parse_bool(             $new_label, 10, $_,"break_label");
	    parse_address(          $new_label, 11, $_,"exec_stmt","statement");

       	    if(!$found) {
                printf STDERR "Unrecognized line, reading a label, line number %d\n", $line_no, $1;
                printf STDERR "  %s\n", $_;
            }
	}	      
	$current_label += 1;	      
    }
    elsif( $_ =~ /^([\w|-]+)[ ]+namespace.(\d+)$/) {
	while(<ARGV>) {
	    $line_no++;
	    if($_ =~ /^$/) {
		last;
	    }
	}
    }
    elsif( $_ =~ /^([\w|-]+)[ ]+class-type-supplement.(\d+)$/) {
	while(<ARGV>) {
	    $line_no++;
	    if($_ =~ /^$/) {
		last;
	    }
	}
    }
    elsif( $_ =~ /Display of IL file .*, produced by the compilation of .*/) {
	$line_no++;
    }
    elsif(!($_ =~ /^\s*$/)) {	
	$line_no++;
	printf STDERR "Unrecognized line, reading outside of any entry, line number %d\n", $line_no, $1;
	printf STDERR "  %s\n", $_;
    }

    # "while(ARGV)" may not fail, so this line helps us get out of the loop
    if(eof ARGV) {
	last;
    }
}

print_region();
print(")\n");
