;; -----------------------------------------------------------------------
 ; Indiana University Opie Lab Software License, Version 1.2
 ;
 ; Copyright (C) 2002 The Trustees of Indiana University.
 ; All rights reserved.
 ;
 ; Redistribution and use in source and binary forms, with or without
 ; modification, are permitted provided that the following conditions are
 ; met:
 ;
 ; 1) All redistributions of source code must retain the above
 ;    copyright notice, the list of authors in the original source
 ;    code, this list of conditions and the disclaimer listed in this
 ;    license;
 ;
 ; 2) All redistributions in binary form must reproduce the above
 ;    copyright notice, this list of conditions and the disclaimer
 ;    listed in this license in the documentation and/or other
 ;    materials provided with the distribution;
 ;
 ; 3) Any documentation included with all redistributions must include
 ;    the following acknowledgement:
 ;
 ;      "This product includes software developed by the Indiana
 ;      University Opie Lab.  For further information please visit
 ;      http://www.cs.indiana.edu/~dswise"
 ;
 ;    Alternatively, this acknowledgment may appear in the software
 ;    itself, and wherever such third-party acknowledgments normally
 ;    appear.
 ;
 ; 4) The name "Indiana University" or "Indiana University
 ;    Opie Lab" shall not be used to endorse or promote
 ;    products derived from this software without prior written
 ;    permission from Indiana University.  For written permission,
 ;    please contact http://www.cs.indiana.edu/~dswise.
 ;
 ; 5) Products derived from this software may not use "Indiana
 ;    Univeristy" name nor may "Indiana University" appear in their name,
 ;    without prior written permission of the Indiana University.
 ;
 ; Indiana University provides no reassurances that the source code
 ; provided does not infringe the patent or any other intellectual
 ; property rights of any other entity.  Indiana University disclaims any
 ; liability to any recipient for claims brought by any other entity
 ; based on infringement of intellectual property rights or otherwise.
 ;
 ; LICENSEE UNDERSTANDS THAT SOFTWARE IS PROVIDED "AS IS" FOR WHICH
 ; NO WARRANTIES AS TO CAPABILITIES OR ACCURACY ARE MADE. INDIANA
 ; UNIVERSITY GIVES NO WARRANTIES AND MAKES NO REPRESENTATION THAT
 ; SOFTWARE IS FREE OF INFRINGEMENT OF THIRD PARTY PATENT, COPYRIGHT, OR
 ; OTHER PROPRIETARY RIGHTS.  INDIANA UNIVERSITY MAKES NO WARRANTIES THAT
 ; SOFTWARE IS FREE FROM "BUGS", "VIRUSES", "TROJAN HORSES", "TRAP
 ; DOORS", "WORMS", OR OTHER HARMFUL CODE.  LICENSEE ASSUMES THE ENTIRE
 ; RISK AS TO THE PERFORMANCE OF SOFTWARE AND/OR ASSOCIATED MATERIALS,
 ; AND TO THE PERFORMANCE AND VALIDITY OF INFORMATION GENERATED USING
 ; SOFTWARE.
;; ------------------------------------------------------------------------

; scil-records.ss
; created by Brooke Chenoweth

; This file uses Chez Scheme records for the different parts of the scil tree

; some of this has been changed from the specifications to match the perl

; As far as I know, all the records in this file match the perl code,
; except for constants, which Steve is planning to change.

; This is the big overall structure
(define-record scil_structure (il_header
			       file_scope  ;file scope memory region
			       fun_mem_region_list
			       ))

(define-record memory_region (number  ; integer  
			      name    ; file scope or name of function 
			      source_file_list   
			      scope_list         
			      pragma_list        
			      type_list          
			      variable_list      
			      routine_list       
			      field_list         
			      dynamic_init_list  
			      param_type_list    
			      constant_list      
			      statement_list     
			      expression_list   
			      label_list
                              matrix_info
                              integer_subscript_list
                              odd_vars
                              even_vars
			      ))

; I've changed the print method to display what the address is linked
; to, if anything. If the output produced isn't helpful, it might be a
; good idea to change print for addresses and/or other records.  If
; you want to suppress the display of ref, call suppress-address-ref
; with an argument of #t
(define-record address
  (mem_region_name 	
   field 
   num)
  ((ref '()))
  ;(print-method
  ;  (lambda (r p wr)
  ;    (display "ref:" p)
  ;    (cond [(eqv? 0 (address-mem_region_name r)) (display "none" p)]
;	    [(or (null? (address-ref r))
	;	 (suppress-address-ref)) (wr (list (address-mem_region_name r)
	;					   (address-field r)
	;					   (address-num r)) p)]
	;    [else (wr (address-ref r) p)])))
  )

(define suppress-address-ref
  (let ([flag #f])
    (lambda x
      (if (null? x)
	  flag
	  (if (boolean? (car x))
	      (set! flag (car x))
	      (error 'suppress-address-ref "Argument ~s is not a boolean value" (car x)))))))


(define-record il_header (primary_source_file            ;source-file address  
			  primary_scope                  ;scope address        
			  main_routine                   ;routine address      
			  compiler_version               ;string               
			  time_of_compilation            ;string               
			  plain_chars_are_signed         ; bool                
			  source_language                ; string              
			  pcc_compatability_mode         ; bool                
			  enum_type_is_integral          ; bool                
			  default_max_member_alignment   ; integer             
			  near_and_far_are_enabled       ; bool                
			  far_data_pointers              ; bool                
			  far_code_pointers              ; bool                
			  UCN_identifiers_used           ; bool                
			  nontag_types_used_in_exception_or_rtti ; type address
			  ))

(define-record source_file (number                    ; source-file number
			    next                      ; source-file address
			    file_name                 ; string 
			    full_name                 ; string 
			    name_as_written           ; string 
			    first_seq_number          ; number 
			    last_seq_number           ; number 
			    first_line_number         ; number 
			    first_child_file          ; source-file address
			    last_child_file           ; source-file address
			    related_file_implicit_include_done ; bool
			    is_include_file                    ; bool	
			    included_by_system_include         ; bool
			    included_by_preinclude             ; bool
			    from_system_include_dir            ; bool
			    ))


(define-record scope (number			; scope number
		      next			; scope address
		      kind			; sck_stuff
		      assoc_block		; statement address
		      lifetime			; lifetime address (lifetimes not supported)
		      constants			; constant address
		      types			; type address
		      variables			; variable address
		      nonstatic_variables	; variable address
		      labels			; label address
		      routines			; routine address
		      asm_entries		; asm-entry address 
		      scopes			; scope address
		      dynamic_init		; dynamic-init address (printed but not supported)
		      pragmas			; pragma address
		      scope_tail        	; kind dependent scope parts
		      ))

; func_prototype or condition
(define-record func_prototype_scope_tail (assoc_type	; type address
					  ))
; function
(define-record function_scope_tail (routine_ptr       	; routine address
				    parameters		; variable address
				    return_value_variable	; variable address
				    ))

; class_struct_union
(define-record class_struct_union_scope_tail (using_decls ; using-decl address
					      ))

(define-record pragma (number		; pragma number
		       next		; pragma address
		       entity		; unknown address
		       position		; number
		       pragma_text	; string
		       ignore_in_backend ; bool
		       kind		; lots of choices
		       ))

(define-record type (number             ; type number
		     next               ; type address
		     kind               ; the type kind
		     size               ; integer
		     alignment          ; integer
		     used_in_exception_or_rtti 	; bool
		     declared_in_function_prototype ; bool
		     is_builtin_va_list ; bool 
		     typeinfo_var       ; variable address
		     name               ; string
		     unmangled_name     ; string
		     decl_position_seq  ; number
		     decl_position_column      	; number
		     referenced         ; bool 
		     needed             ; bool
		     is_local_to_function ; bool
		     name_linkage              	; no|C|C++
		     has_associated_pragma      ; bool
		     static_used_by_instantiation ; bool
		     reference                 	; type address
		     pointer                   	; type address
		     qualified                 	; type address
		     type_tail ; a record of the kind dependent type parts
		     ))

(define-record int_type_tail (int_kind                  
			      explicitly_signed         ; bool
			      enum_info_constant_list   ; constant address
			      enum_info_affiliated_type ; type address
			      bool_type                 ; bool
			      ))

(define-record float_type_tail (float_kind
				))

(define-record pointer_type_tail (type_pointed_to       ; type address
				  is_reference          ; bool
				  ))

(define-record routine_type_tail (return_type                   ; type address
				  param_type_list               ; param-type address
				  assoc_routine                 ; routine address
				  has_ellipsis                  ; bool
				  prototyped                    ; bool
				  routine_name_linkage          ; no|internal|C|C++
				  routine_name_linkage_is_explicit   	; bool
				  lint_argsused_flag            ; bool
				  value_returned_by_cctor       ; bool
				  assoc_routine_is_ctor         ; bool
				  assoc_routine_is_dtor         ; bool
				  lint_varargs_count            ; integer
				  arg_pragma                    ;    
				  qualifiers                    ;    
				  prototype_scope               ; scope address
				  ))

(define-record array_type_tail (element_type            ; type address
				is_static               ; bool
				is_variable_size_array  ; bool
				is_vla                  ; bool
				has_assoc_vla_dimension ; bool
				element_count_expr      ; expression address
				number_of_elements      ; integer
				bound_constant          ; constant address
				))

(define-record struct_type_tail (field_list             ; field address
				 any_const_member       ; bool
				 any_mutable_member     ; bool
				 abstract               ; bool
				 definition_needed      ; bool
				 max_member_alignment   ; integer
				 ))

(define-record typeref_type_tail (type_ref_type         ; type address          
				  qualifiers            ; const|restrict|__far|__near|__unaligned|??
				  explicit_memory_attribute_made_implicit       ; bool
				  ))

(define-record ptr_to_member_type_tail (class_of_which_a_member  ; type address
					))

(define-record variable (number                 ; variable number
			 next                   ; variable address
			 name                   ; string
			 unmangled_name         ; string
			 decl_position_seq      ; integer
			 decl_position_column   ; integer
			 referenced             ; bool  
			 needed                 ; bool  
			 is_local_to_function   ; bool  
			 name_linkage           ; no|C|C++
			 has_associated_pragma  ; bool
			 static_used_by_instantiation ; bool
			 type                   ; type address
			 assoc_param_type       ; param-type address
			 storage_class          ; sc_stuff
			 declared_storage_class ; sc_stuff
			 init_kind              ; (initk_)static|dynamic|zero|function_local
			 address_taken          ; bool
			 dynamic                ; dynamic-init address
			 param_value_has_been_changed    ; bool
			 param_used_more_than_once       ; bool
			 is_parameter           ; bool  
			 is_handler_param       ; bool
			 referenced_non_locally ; bool
			 is_partially_initialized ; bool
			 superseded_external    ; bool
			 is_compound_literal    ; bool
			 has_parenthesized_initializer ; bool
			 declared_type          ; type address
                         constant               ; constant address
			 even_shadow
			 odd_shadow
			 register
			 ))

(define-record field (number                            ; field number
		      next                              ; field address
		      name                              ; string
		      decl_position_seq                 ; number
		      decl_position_column              ; number
		      is_class_member                   ; bool (C++)
		      parent_class_type                 ; type address      
		      referenced                        ; bool
		      needed                            ; bool
		      name_linkage                      ; no|internal|C|C++
		      type                              ; type address
		      offset                            ; number
		      assoc_type                        ; type address
		      ))

; changed to match the perl
(define-record param_type (number                       ; param-type number
  			   next                         ; param-type address
			   type                         ; type address
			   dilated_type                 ;
	; 		   name                         ; string 
; 			   default_arg_expr             ; expression address                         
; 			   qualifiers                   ; const|restrict|__far|__near|__unaligned|???
			   ))

; changed big time to match the perl
; gave up and stopped handling most fields
;(define-record constant (number                         ; constant number
;			 next                           ; constant address
;			 type                           ; type address
;			 kind                           ; ck_stuff
	; 		 length		                ; number
; 			 value		                ; string
; 			 integer_value		        ; number
; 			 implicit_cast                  ; bool
; 			 address_kind                   ; (abk_)routine|variable|constant|uuidof 
; 			 address_offset                 ; ?
; 			 constant	                ; constant address
; 			 referenced                     ; bool
; 			 needed                         ; bool
; 			 non_arithmetic                 ; bool
; 			 is_simple_zero                 ; bool
; 			 variable                       ; variable address
; 			 float_value		        ; number
; 			 null_pointer_constant_ruled_out  ; bool
;			 ))

(define-record constant (number				; constant number
			 next				; constant address
			 name				; string
			 unmangled_name		        ; string
			 decl_position_seq		; number
			 decl_position_column		; number
			 referenced			; bool
			 needed			        ; bool
			 is_local_to_function		; bool
			 name_linkage			; no|internal|C|C++
			 has_associated_pragma		; bool
			 static_used_by_instantiation	; bool
			 type				; type address
			 expr				; expr address (only if RECORD_CONSTANT_EXPRESSIONS_IN_IL)
			 implicit_cast			; bool
			 explicit_cast_applied		; bool
			 is_reinterpret_cast		; bool
			 non_arithmetic			; bool
			 is_simple_zero			; bool
			 null_pointer_constant_ruled_out	; bool
			 kind				; ck_stuff
			 constant_tail              ; kind specific constant parts
			 ))

(define-record integer_constant_tail (integer_value	       ; number
				      ))

(define-record string_constant_tail (length			;	number
				     value			;	string
				     ))

(define-record float_constant_tail (float_value		;	number
				    ))

(define-record address_constant_tail (address_kind    ; (abk_)routine|variable|constant|uuidof
				      address_offset
				      address         ; address of a routine, variable, constant, or type
				      ))

(define-record aggregate_constant_tail (first_constant		;	constant address
					last_constant		;	constant address
					))

; changed to match perl
(define-record expression (number                           ; expression number
			   next                             ; expression address
			   kind                             ; enk_stuff
			   operation_kind                   ; eok_stuff
			   routine                          ; routine address
			   type                             ; type address
			   operands                         ; expression address
			   variable                         ; variable address
			   constant                         ; constant address
			   result_is_not_used               ; bool
			   compiler_generated               ; bool
			   field	                    ; field address
                           odd_typing                       
                           even_typing
                           odd_terms
                           even_terms
			   updated                          ; has it been updated yet
			   ))


; big changes to match the perl
(define-record statement (number                ; statement number
			  next                  ; statement address
			  kind                  ; stmk_stuff
			  position              ; number
			  final_position
			  expr	                ; expression address
			  statements     	; address
			  assoc_scope  		; address
			  lifetime   		; address
			  end_of_block_reachable; bool
			  then_statement      	; statement address
			  else_statement      	; statement address
			  initialization     	; statement address
			  statement          	; statement address
			  increment          	; expression address
			  has_empty_else_clause ; bool
			  dynamic_init		; dynamic-init address
			  label                 ; label address
			  deleted		; bool -- used by compiler
			  ))



; I got this from Steve's perl file and so was guessing a bit.
; I just tried to make it similar to the others
(define-record routine (number                            ; number
			next                              ; address
			name                              ; string
			unmangled_name                    ; string
			decl_position_seq                 ; number
			decl_position_column              ; number
			referenced                        ; bool
			needed                            ; bool
			is_local_to_function              ; bool
			name_linkage                      ; no|internal|C|C++
			type                              ; type address
			storage_class                     ; ??
			defined                           ; bool
			is_called                         ; bool
			definition_needed                 ; bool
			assoc_scope                       ; this is the number of the associated memory region 
			compiler_generated                ; bool
			called                            ; bool
                        param_typing
			))

(define-record label (number                              ; number
		      next                                ; label address
		      name                                ; string
		      decl_position_seq                   ; number
		      decl_position_column                ; number
		      referenced                          ; bool
		      needed                              ; bool
		      is_local_to_function                ; bool
		      name_linkage                        ; no|internal|C|C++
		      reachable_by_fall_through           ; bool
		      break_label                         ; bool
                      exec_stmt                           ; statement address
		      ))

(define-record dynamic_init (number                       ; number
			     next                         ; dynamic_init address
			     kind                         ; symbol
			     variable                     ; variable address
			     constant                     ; constant address
			     ))
			     
			     

(define-record one_matrix (variable
                           stride
                           deref_exprs
                           even_vars
                           odd_vars
                          )
)

     				      
